/*
    Copyright (C) 2011 William Hart
    Copyright (C) 2012 Sebastian Pancratz

    This file is part of FLINT.

    FLINT is free software: you can redistribute it and/or modify it under
    the terms of the GNU Lesser General Public License (LGPL) as published
    by the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.  See <https://www.gnu.org/licenses/>.
*/

#include "fmpz.h"
#include "fmpz_vec.h"
#include "fmpz_mod.h"
#include "fmpz_mod_poly.h"

slong _fmpz_mod_poly_xgcd_euclidean_f(fmpz_t f, fmpz *G, fmpz *S, fmpz *T,
                                   const fmpz *A, slong lenA,
                                   const fmpz *B, slong lenB,
                                   const fmpz_t invB, const fmpz_mod_ctx_t ctx)
{
    _fmpz_vec_zero(G, lenB);
    _fmpz_vec_zero(S, lenB - 1);
    _fmpz_vec_zero(T, lenA - 1);

    if (lenB == 1)
    {
        fmpz_set_ui(f, 1);
        fmpz_set(G + 0, B + 0);
        fmpz_one(T + 0);
        return 1;
    }
    else
    {
        fmpz *Q, *R;
        slong lenQ, lenR, lenD = 0;
        TMP_INIT;
        TMP_START;

        FMPZ_VEC_TMP_INIT(Q, 2*lenA);
        R = Q + lenA;

        _fmpz_mod_poly_divrem_f(f, Q, R, A, lenA, B, lenB, ctx);
        if (!fmpz_is_one(f))
            goto cleanup2;

        lenR = lenB - 1;
        FMPZ_VEC_NORM(R, lenR);

        if (lenR == 0)
        {
            fmpz_set_ui(f, 1);
            _fmpz_vec_set(G, B, lenB);
            fmpz_one(T + 0);

            FMPZ_VEC_TMP_CLEAR(Q, 2*lenA);

            TMP_END;
            return lenB;
        }
        else
        {
            fmpz_t inv;
            fmpz *D, *U, *V1, *V3, *W;
            slong lenU, lenV1, lenV3, lenW;

            fmpz_init(inv);
            FMPZ_VEC_TMP_INIT(W, FLINT_MAX(5*lenB, lenA + lenB));
            D  = W  + lenB;
            U  = D  + lenB;
            V1 = U  + lenB;
            V3 = V1 + lenB;

            lenU = 0;
            _fmpz_vec_set(D, B, lenB);
            lenD = lenB;
            fmpz_one(V1 + 0);
            lenV1 = 1;
            lenV3 = 0;
            FMPZ_VEC_SWAP(V3, lenV3, R, lenR);

            do {
                fmpz_gcdinv(f, inv, V3 + (lenV3 - 1), fmpz_mod_ctx_modulus(ctx));
                if (!fmpz_is_one(f))
                    goto cleanup;

                _fmpz_mod_poly_divrem_basecase(Q, D, D, lenD, V3, lenV3, inv, ctx);
                lenQ = lenD - lenV3 + 1;
                lenD = lenV3 - 1;
                FMPZ_VEC_NORM(D, lenD);

                if (lenV1 >= lenQ)
                    _fmpz_mod_poly_mul(W, V1, lenV1, Q, lenQ, ctx);
                else
                    _fmpz_mod_poly_mul(W, Q, lenQ, V1, lenV1, ctx);
                lenW = lenQ + lenV1 - 1;

                _fmpz_mod_poly_sub(U, U, lenU, W, lenW, ctx);
                lenU = FLINT_MAX(lenU, lenW);
                FMPZ_VEC_NORM(U, lenU);

                FMPZ_VEC_SWAP(U, lenU, V1, lenV1);
                FMPZ_VEC_SWAP(D, lenD, V3, lenV3);

            } while (lenV3 != 0);

            _fmpz_vec_set(G, D, lenD);
            _fmpz_vec_set(S, U, lenU);
            {
                lenQ = lenA + lenU - 1;

                _fmpz_mod_poly_mul(Q, A, lenA, S, lenU, ctx);
                _fmpz_mod_poly_neg(Q, Q, lenQ, ctx);
                _fmpz_mod_poly_add(Q, G, lenD, Q, lenQ, ctx);

                _fmpz_mod_poly_divrem(T, W, Q, lenQ, B, lenB, invB, ctx);
            }

cleanup:
            fmpz_clear(inv);
            FMPZ_VEC_TMP_CLEAR(W, FLINT_MAX(5*lenB, lenA + lenB));

cleanup2:
            FMPZ_VEC_TMP_CLEAR(Q, 2*lenA);

            TMP_END;
            return lenD;
        }
    }
}

void
fmpz_mod_poly_xgcd_euclidean_f(fmpz_t f, fmpz_mod_poly_t G, fmpz_mod_poly_t S,
          fmpz_mod_poly_t T, const fmpz_mod_poly_t A, const fmpz_mod_poly_t B,
                                                      const fmpz_mod_ctx_t ctx)
{
    if (A->length < B->length)
    {
        fmpz_mod_poly_xgcd_euclidean_f(f, G, T, S, B, A, ctx);
    }
    else  /* lenA >= lenB >= 0 */
    {
        const fmpz * p = fmpz_mod_ctx_modulus(ctx);
        const slong lenA = A->length, lenB = B->length;
        fmpz_t inv;

        fmpz_init(inv);
        if (lenA == 0)  /* lenA = lenB = 0 */
        {
            fmpz_set_ui(f, 1);
            fmpz_mod_poly_zero(G, ctx);
            fmpz_mod_poly_zero(S, ctx);
            fmpz_mod_poly_zero(T, ctx);
        }
        else if (lenB == 0)  /* lenA > lenB = 0 */
        {
            fmpz_gcdinv(f, inv, fmpz_mod_poly_lead(A, ctx), p);
            fmpz_mod_poly_scalar_mul_fmpz(G, A, inv, ctx);
            fmpz_mod_poly_zero(T, ctx);
            fmpz_mod_poly_set_fmpz(S, inv, ctx);
        }
        else  /* lenA >= lenB >= 1 */
        {
            fmpz *g, *s, *t;
            slong lenG;

            fmpz_gcdinv(f, inv, fmpz_mod_poly_lead(B, ctx), p);

            if (!fmpz_is_one(f))
               goto cleanup;

            if (G == A || G == B)
            {
                g = _fmpz_vec_init(FLINT_MIN(lenA, lenB));
            }
            else
            {
                fmpz_mod_poly_fit_length(G, FLINT_MIN(lenA, lenB), ctx);
                g = G->coeffs;
            }
            if (S == A || S == B)
            {
                s = _fmpz_vec_init(lenB);
            }
            else
            {
                fmpz_mod_poly_fit_length(S, lenB, ctx);
                s = S->coeffs;
            }
            if (T == A || T == B)
            {
                t = _fmpz_vec_init(lenA);
            }
            else
            {
                fmpz_mod_poly_fit_length(T, lenA, ctx);
                t = T->coeffs;
            }

            lenG = _fmpz_mod_poly_xgcd_euclidean_f(f, g, s, t,
                                     A->coeffs, lenA, B->coeffs, lenB, inv, ctx);

            if (G == A || G == B)
            {
                _fmpz_vec_clear(G->coeffs, G->alloc);
                G->coeffs = g;
                G->alloc  = FLINT_MIN(lenA, lenB);
            }
            if (S == A || S == B)
            {
                _fmpz_vec_clear(S->coeffs, S->alloc);
                S->coeffs = s;
                S->alloc  = lenB;
            }
            if (T == A || T == B)
            {
                _fmpz_vec_clear(T->coeffs, T->alloc);
                T->coeffs = t;
                T->alloc  = lenA;
            }

            if (!fmpz_is_one(f))
               goto cleanup;

            _fmpz_mod_poly_set_length(G, lenG);
            _fmpz_mod_poly_set_length(S, FLINT_MAX(lenB - lenG, 1));
            _fmpz_mod_poly_set_length(T, FLINT_MAX(lenA - lenG, 1));
            _fmpz_mod_poly_normalise(S);
            _fmpz_mod_poly_normalise(T);

            if (!fmpz_is_one(fmpz_mod_poly_lead(G, ctx)))
            {
                fmpz_gcdinv(f, inv, fmpz_mod_poly_lead(G, ctx), p);
                fmpz_mod_poly_scalar_mul_fmpz(G, G, inv, ctx);
                fmpz_mod_poly_scalar_mul_fmpz(S, S, inv, ctx);
                fmpz_mod_poly_scalar_mul_fmpz(T, T, inv, ctx);
            }
        }

cleanup:

        fmpz_clear(inv);
    }
}
