@import url("pygments.css");
@import url("classic.css");

body { font-family: "arial", sans-serif; }

div .toctree-wrapper { column-count: 2; }

div .toctree-wrapper > ul { margin: 0; }

ul .toctree-l1 { 
    margin: 0;
    -webkit-column-break-inside: avoid;
    page-break-inside: avoid;
    break-inside: avoid-column;
}

div.body {
    min-width: 450px;
    max-width: 100%;
}

li p { margin: 0.2em; }

