/*
    Copyright (C) 2020 Fredrik Johansson

    This file is part of FLINT.

    FLINT is free software: you can redistribute it and/or modify it under
    the terms of the GNU Lesser General Public License (LGPL) as published
    by the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.  See <https://www.gnu.org/licenses/>.
*/

#include "test_helpers.h"
#include "ca.h"

TEST_FUNCTION_START(ca_ctx_init_clear, state)
{
    ca_ctx_t ctx;

    ca_ctx_init(ctx);
    ca_ctx_clear(ctx);

    TEST_FUNCTION_END(state);
}
